#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
from dataclasses import dataclass
from typing import ClassVar

from http_tools import ContentType

from .body_parameter import BodyParameter


@dataclass(frozen=True, kw_only=True)
class JsonWholeBodyParameter(BodyParameter, abc.ABC):
    body_mime_type: ClassVar[ContentType] = ContentType.Json

    def __init__(self):
        raise NotImplementedError("requires inheritance restructuration")



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
# import typing
from dataclasses import dataclass, field
from functools import cached_property
from typing import Generic, ClassVar, Mapping

from frozendict import frozendict
from http_tools import IncomingRequest
from init_helpers import raise_if

from ..exceptions import Unauthorized
from .security_scheme import SecurityScheme, AuthInfo, AuthToken
from .security_scheme_type import SecuritySchemeType
from ... import SpecRef, SpecResource


@dataclass(frozen=True, slots=True, repr=False)
class HttpSecurityScheme(SecurityScheme, Generic[AuthInfo], abc.ABC):
    type_: ClassVar[SecuritySchemeType] = SecuritySchemeType.http
    header_scheme: ClassVar[str]

    def get_spec_dict(self, dependency_to_ref: Mapping[SpecResource, SpecRef]) -> frozendict:
        return SecurityScheme.get_spec_dict(self, dependency_to_ref) | {'scheme': self.header_scheme.strip().lower()}

    def get_key(self) -> str:
        return f'{self.type_.name}_{self.header_scheme}'.lower().strip()

    async def _extract_token(self, incoming_request: IncomingRequest) -> AuthToken:
        dirty_token: str = incoming_request.metadata.header_name_to_value.get("Authorization")
        raise_if(not isinstance(dirty_token, str), Unauthorized("Failed to get token from Authorization header"))
        raise_if(not dirty_token.startswith(self.header_scheme), Unauthorized("Wrong authorization scheme"))
        return dirty_token.removeprefix(self.header_scheme)


@dataclass(frozen=True, repr=False)
class HttpBearerSecurityScheme(HttpSecurityScheme, Generic[AuthInfo]):
    header_scheme: ClassVar[str] = 'Bearer '


@dataclass(frozen=True, repr=False)
class HttpBasicSecurityScheme(HttpSecurityScheme, Generic[AuthInfo]):
    header_scheme: ClassVar[str] = 'Basic '



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .endpoint_body import EndpointBody
from .gen_endpoint import gen_query_endpoint, gen_json_endpoint, get_securities_arg_names
from .openapi_server import OpenApiServer
from .endpoint import Endpoint



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
# import functools
from dataclasses import dataclass
from typing import ClassVar

from .spec_ref import SpecRef
from .spec_resource import SpecResource


@dataclass(frozen=True, slots=True)
class MaybeReferencableResource(SpecResource):
    can_be_inlined: ClassVar[bool] = True

    @property
    @abc.abstractmethod
    def spec_path(cls) -> tuple[str, ...]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_key(self) -> str | None:
        raise NotImplementedError

    def get_spec_ref(self, key: str = '') -> SpecRef | str:
        return SpecRef(f'#/{"/".join(self.spec_path)}/{key if key else self.get_key()}')

    # def get_spec_ref(self) -> SpecRef | str:
    #     return SpecRef(f'{"/".join(self.spec_path + [self.get_key()])}')

    # @functools.cache
    # def get_spec_referencable_dependencies(self) -> frozenset['ReferencableResource']:
    #     return frozenset(filter(lambda x: isinstance(x, ReferencableResource), self.get_spec_dependencies()))



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .api_key_security_scheme import ApiKeySecurityScheme
from .http_security_scheme import HttpSecurityScheme, HttpBearerSecurityScheme, HttpBasicSecurityScheme
from .oauth2_flow import OAuth2Flow
from .oauth2_security_scheme import OAuth2SecurityScheme
from .openId_connect_security_scheme import OpenIdConnectSecurityScheme
from .security_scheme import SecurityScheme, AuthInfo, AuthToken, resolver_cache
from .security_scheme_type import SecuritySchemeType



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
class SecurityException(Exception):
    pass


class Unauthorized(SecurityException):
    pass


class Forbidden(SecurityException):
    pass


class AuthInfoException(SecurityException):
    pass



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass, field
from typing import ClassVar, Any, Iterable

from dict_caster.extras import is_iterable
from dynamic_types import create_type, prepare_class_name
from http_tools import ContentType, IncomingRequest

from .body_parameter import BodyParameter


@dataclass(frozen=True, kw_only=True)
class RawBodyParameter(BodyParameter):
    body_mime_type: ClassVar[ContentType | tuple[ContentType, ...]] = ContentType.Octet
    name: str = field(init=False, default=None)
    schema: type = field(init=False, default=bytes)

    def _get(self, incoming_request: IncomingRequest) -> Any:
        return incoming_request.payload

    def __class_getitem__(cls, content_type: ContentType | Iterable[ContentType]) -> type:
        if is_iterable(content_type):
            content_type = tuple(content_type)
        return create_type(prepare_class_name([cls], content_type), [cls], {'body_mime_type': content_type})



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses
import functools
import logging
import re
import types
from decimal import Decimal
from enum import Enum
from types import UnionType
from typing import Callable, Any, get_args, get_origin, get_type_hints, Union, Literal

from frozendict import frozendict
from more_itertools import first
from http_tools import WrappedAnswerBody
from init_helpers import try_extract_type_notes, DataclassProtocol, raise_if, Jsonable, AnyType
from init_helpers import NoValue

from .any_of_schema import AnyOfSchema
from .all_of_schema import AllOfSchema
from .array_schema import ArraySchema
from .base_schema import BaseSchema
from .boolean_schema import BooleanSchema
from .empty_schema import EmptySchema
from .integer_schema import IntegerSchema
from .null_schema import NullSchema
from .number_schema import NumberSchema
from .object_schema import ObjectSchema
from .schema_format import StringSchemaFormat
from .string_schema import StringSchema
from .type_schema import TypeSchema
from ..example import BaseExample, Example
from ..spec_ref import SpecRef
from ...server.endpoint_body import EndpointBody

# if TYPE_CHECKING:
#     from spec_schema import SpecSchema

logger = logging.getLogger(__name__)
#
# @functools.cache
# def build_spec_schema(schema: AnyType) -> 'SpecResource | NamedSpecSchema':
#     from spec_schema import SpecSchema
#     from named_spec_schema import NamedSpecSchema
#
#     if isinstance(schema, SpecSchema):
#         return schema
#
#     clean_type, notes = try_extract_type_notes(schema)
#
#     if examples := tuple(note for note in notes if isinstance(note, BaseExample)):
#         example = examples[0].value
#
#     if dataclasses.is_dataclass(schema):
#         return NamedSpecSchema(type_=schema, name=f"schema_{clean_type.__name__}")
#
#     return SpecSchema(schema)


class SchemaBuilder:
    dataclass_to_schema: dict[DataclassProtocol, BaseSchema | SpecRef] = {}
    dataclass_to_ref: dict[DataclassProtocol, SpecRef] = {}

    @classmethod
    def build_schema(
            cls,
            schema: AnyType, default: Any = NoValue,
            description: str = '',
            examples: tuple[BaseExample, ...] = tuple(),
            enum: tuple[Jsonable, ...] | None = None,
    ) -> BaseSchema:
        logger.debug('_build_schema: %s', schema)
        schema, notes = try_extract_type_notes(schema)
        examples = (examples or tuple()) + tuple(note for note in notes if isinstance(note, Example))
        pattern = first((note.pattern for note in notes if isinstance(note, re.Pattern)), default=None)
        description: str = description or first((note for note in notes if isinstance(note, str)), default='')

        kwargs = {
            "description": description, "examples": examples, "default": default, "enum": enum, "pattern": pattern}

        if schema is Any:
            schema = dict | list | str | float | int | bool | None

        if origin_type := get_origin(schema):
            result = cls._get_generic_description(origin_type, get_args(schema), **kwargs)
        elif schema in (dict, frozendict, list, set, frozenset, tuple):
            result = cls._get_generic_description(schema, **kwargs)
        elif dataclasses.is_dataclass(schema):
            result = cls._get_dataclass_schema(schema, **kwargs)
        # elif isinstance(schema, type) and issubclass(schema, Enum):
        #     result = cls._get_primitive_description(schema)

        else:
            result = cls._get_primitive_description(schema, **kwargs)
        #
        # # AVOID MODIFICATION OF CACHED VALUES!!!
        # if isinstance(schema, type) and issubclass(schema, Enum):
        #     result = result | {'enum': [e.value for e in schema]}
        # if default is None or isinstance(default, (int, float, str, list, dict)):
        #     result = result | {'default': default}
        # if notes:
        #     if examples := tuple(note for note in notes if isinstance(note, Example)):
        #         result = result | {'example': examples[0].value}
        #     if descriptions := [note for note in notes if isinstance(note, str)]:
        #         result = result | {'description': descriptions[0]}
        # # if default == NoValue:  # TODO: think about it, should we place "required" inside schema properties or not?
        # #     description['required'] = True
        return result

    @classmethod
    def _get_primitive_description(
            cls, type_: type,
            description: str = '',
            examples: tuple[BaseExample, ...] = tuple(),
            default: Any = NoValue,
            enum: tuple[Jsonable, ...] | None = None,
            pattern: str | None = None,
    ) -> TypeSchema:
        extra_kwargs: dict[str, Any] = {'enum': enum} if enum else {}
        if not enum and issubclass(type_, Enum):
            extra_kwargs['enum'] = tuple(e.value for e in type_)
            extra_kwargs['key'] = type_.__name__
        extra_kwargs |= {'pattern': pattern} if pattern else {}

        for primitive_type, factory in cls.primitive_type_to_factory.items():
            if issubclass(type_, primitive_type):
                return factory(description=description, examples=examples, default=default, **extra_kwargs)
        raise TypeError(f"Unknown type: {type_}")

    @classmethod
    def _generate_dataclass_schema(cls, schema: type[DataclassProtocol]) -> ObjectSchema | SpecRef:
        if ref := cls.dataclass_to_ref.get(schema):
            return ref
        if key := None if issubclass(schema, WrappedAnswerBody | EndpointBody) else schema.__name__:
            cls.dataclass_to_ref[schema] = ObjectSchema(key=key).get_spec_ref()
        key_to_type_hint = get_type_hints(schema)
        key_to_schema = {}
        for field_ in dataclasses.fields(schema):
            if not field_.repr:
                continue
            try:
                field_schema = cls.build_schema(field_.type)
            except TypeError as e:
                field_schema = cls.build_schema(key_to_type_hint[field_.name])
            if field_.default is not dataclasses.MISSING:
                if not field_schema.has_default and not field_schema.key:
                    field_schema = dataclasses.replace(field_schema, default=field_.default)
                else:
                    field_schema = AllOfSchema([field_schema, EmptySchema(default=field_.default)])
            key_to_schema[field_.name] = field_schema
        result = ObjectSchema(item_key_to_schema=key_to_schema, key=key)
        return result

    @classmethod
    def _get_dataclass_schema(
            cls,
            schema: type[DataclassProtocol],
            description: str = '',
            examples: tuple[BaseExample, ...] | None = None,
            default: Any = NoValue,
            enum: tuple[Jsonable, ...] | None = None,
            pattern: str | None = None,
    ) -> ObjectSchema | SpecRef:
        # noinspection PyTypeChecker
        result = cls.dataclass_to_schema.get(schema) or cls._generate_dataclass_schema(schema)
        cls.dataclass_to_schema[schema] = result
        if description or examples or default is not NoValue:
            result = AllOfSchema([result, EmptySchema(description=description, examples=examples, default=default)])
        return result

    @classmethod
    def _get_generic_description(
            cls, origin_type: type | UnionType, type_args: tuple[type | UnionType, ...] | None = None,
            description: str = '',
            examples: tuple[BaseExample, ...] = tuple(),
            default: Any = NoValue,
            enum: tuple[Jsonable, ...] | None = None,
            pattern: str | None = None,
    ) -> BaseSchema:
        kwargs = {'description': description, 'examples': examples, 'default': default}
        if origin_type in (Union, types.UnionType):
            raise_if(not type_args, TypeError(f'bad {origin_type.__name__} {type_args=}'))
            schema_to_subtype = {cls.build_schema(subtype): subtype for subtype in type_args}
            return AnyOfSchema(
                tuple(schema_to_subtype), **kwargs) if len(schema_to_subtype) > 1 else first(schema_to_subtype)
        elif origin_type in (dict, frozendict):
            raise_if(type_args and len(type_args) != 2, TypeError(f'bad {origin_type.__name__} {type_args=}'))
            return ObjectSchema(additional_items_schema=cls.build_schema(type_args[1]) if type_args else None, **kwargs)
        elif origin_type in (list, set, frozenset):
            raise_if(type_args and len(type_args) != 1, TypeError(f'bad {origin_type.__name__} {type_args=}'))
            return ArraySchema(items_schema=cls.build_schema(type_args[0]) if type_args else None,
                               is_uniqueness_required=origin_type != list, **kwargs)
        elif origin_type is tuple:
            items_types = {t for t in type_args or () if t != Ellipsis}
            # TODO: limit items amount if Ellipsis is missing
            # unbounded = Ellipsis in type_args
            raise_if(len(items_types) > 1, TypeError(f'Inconsistent tuple got: {origin_type.__name__} {type_args=}'))
            return ArraySchema(items_schema=cls.build_schema(first(items_types)) if items_types else None, **kwargs)
        elif origin_type is Literal:
            raise_if(not type_args, TypeError(f'bad {origin_type.__name__} {type_args=}'))
            kwargs['enum'] = type_args
            arg_types = {type(arg) for arg in type_args}
            if len(arg_types) > 1:
                return AnyOfSchema(tuple(cls.build_schema(subtype) for subtype in arg_types), **kwargs)
            else:
                return cls.build_schema(first(arg_types), **kwargs)

        raise TypeError(f"Unknown origin_type: {origin_type}")

    # @classmethod
    # def _get_union_description(
    #         cls, type_args: tuple[type | UnionType] | None, description: str | None = None,
    #         examples: tuple[BaseExample, ...] | None = None, default: Any = NoValue
    # ) -> AnyOfSchema:
    #     raise_if(type_args is not None and len(type_args) != 2, TypeError(f'bad dict item type: {type_args}'))
    #     return AnyOfSchema(
    #         tuple(cls.build_schema(subtype) for subtype in type_args), description=description,
    #         examples=examples, default=default)
    #
    # @classmethod
    # def _get_dict_description(
    #         cls, type_args: tuple[type | UnionType] | None, description: str | None = None,
    #         examples: tuple[BaseExample, ...] | None = None, default: Any = NoValue
    # ) -> ObjectSchema:
    #     raise_if(type_args is not None and len(type_args) != 2, TypeError(f'bad dict item type: {type_args}'))
    #     return ObjectSchema(
    #         additional_items_schema=cls.build_schema(type_args[1]) if type_args else None, description=description,
    #         examples=examples, default=default)
    #
    # @classmethod
    # def _get_list_description(
    #         cls, type_args: tuple[type | UnionType] | None, description: str | None = None,
    #         examples: tuple[BaseExample, ...] | None = None, default: Any = NoValue
    # ) -> ArraySchema:
    #     raise_if(type_args is not None and len(type_args) != 1, TypeError(f'bad list item type: {type_args}'))
    #     return ArraySchema(
    #         items_schema=cls.build_schema(type_args[0]) if type_args else None,
    #         description=description, examples=examples, default=default)
    #
    # @classmethod
    # def _get_set_description(
    #         cls, type_args: tuple[type | UnionType] | None, description: str | None = None,
    #         examples: tuple[BaseExample, ...] | None = None, default: Any = NoValue
    # ) -> ArraySchema:
    #     raise_if(type_args is not None and len(type_args) != 1, TypeError(f'bad set item type: {type_args}'))
    #     return ArraySchema(
    #         items_schema=cls.build_schema(type_args[0]) if type_args else None, is_uniqueness_required=True,
    #         description=description, examples=examples, default=default)
    #
    # origin_type_to_factory: ClassVar[dict[type | UnionType, Callable[[tuple[type | UnionType, ...]], TypeSchema]]] = {
    #     Union: _get_union_description,
    #     types.UnionType: _get_union_description,
    #     dict: _get_dict_description,
    #     frozendict.frozendict: _get_dict_description,
    #     list: _get_list_description,
    #     set: _get_set_description,
    #     frozenset: _get_set_description,
    # }

    primitive_type_to_factory: dict[type, Callable[..., TypeSchema]] = {
        bool: BooleanSchema,  # MUST be before int
        int: IntegerSchema,
        float: NumberSchema,
        bytes: functools.partial(StringSchema, format_=StringSchemaFormat.binary),
        bytearray: functools.partial(StringSchema, format_=StringSchemaFormat.binary),
        str: StringSchema,
        Decimal: functools.partial(StringSchema, pattern=r'^\d*\.?\d*$'),
        type(None): NullSchema,
    }



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from dataclasses import dataclass, field
from typing import Mapping, ClassVar

from frozendict import frozendict
from init_helpers import Jsonable

from openapi_tools.spec.spec_resource import SpecResource, SpecRef
from .base_schema import BaseSchema
from .type_schema import TypeSchema


@dataclass(frozen=True, slots=True)
class ObjectSchema(TypeSchema):
    type_: ClassVar[str] = "object"
    item_key_to_schema: Mapping[str, BaseSchema | SpecRef] = field(default_factory=frozendict)
    additional_items_schema: BaseSchema | SpecRef | None = None

    def __post_init__(self):
        object.__setattr__(self, 'item_key_to_schema', frozendict(self.item_key_to_schema))

    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        result = set()
        if self.item_key_to_schema:
            result = result.union({v for v in self.item_key_to_schema.values() if isinstance(v, SpecResource)})
        if isinstance(self.additional_items_schema, SpecResource):
            result.add(self.additional_items_schema)
        return frozenset(BaseSchema.get_spec_dependencies(self) | result)

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef | dict]) -> frozendict[str, Jsonable]:
        result = {
            'additionalProperties':
                dependency_to_ref[additional_items_schema]
                if isinstance((additional_items_schema := self.additional_items_schema), SpecResource)
                else False
        }
        if self.item_key_to_schema is not None:
            if required := [key for key, value in self.item_key_to_schema.items() if not value.has_default]:
                result['required'] = required
            result['properties'] = {
                key: dependency_to_ref[value] if isinstance(value, SpecResource) else value
                for key, value in self.item_key_to_schema.items()
            }
        return frozendict(TypeSchema.get_spec_dict(self, dependency_to_ref) | result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = list(TypeSchema._get_repr_parts(self))
        if self.item_key_to_schema:
            parts.append(f'item_key_to_schema={dict(self.item_key_to_schema)}')
        parts += [f'additional_items_schema={self.additional_items_schema!r}'] if self.additional_items_schema else []
        return tuple(parts)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from enum import StrEnum


class OAuth2Flow(StrEnum):
    implicit = "implicit"
    authorizationCode = "authorizationCode"
    password = "password"
    clientCredentials = "clientCredentials"



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from dataclasses import dataclass
from typing import ClassVar, Mapping

from frozendict import frozendict
from init_helpers import raise_if, Jsonable

from .spec_ref import SpecRef
from .referencable_resource import ReferencableResource, SpecResource


@dataclass(frozen=True, slots=True)
class Tag(ReferencableResource):
    can_be_inlined: ClassVar[bool] = False
    spec_path: ClassVar[list[str]] = ['tags']
    name: str
    description: str = ''
    docs_url: str = ''
    docs_description: str = ''

    def __post_init__(self):
        raise_if(not self.name, ValueError('Tag name nust be non empty string'))
        raise_if(self.docs_description and not self.docs_url, ValueError('Got docs_description without docs url'))

    @functools.cache
    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        # print('get_spec_dependencies', self)
        return frozenset()

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result = {'name': self.name}
        if description := self.description:
            result['description'] = description
        if docs_url := self.docs_url:
            docs = result['externalDocs'] = {'url': docs_url}
            if docs_description := self.docs_description:
                docs['description'] = docs_description
        return frozendict(result)

    def get_spec_ref(self, key: str = '') -> str:
        return self.name

    def get_key(self) -> str:
        return self.name

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = [f'{self.name!r}']
        parts += [f'description={self.description!r}'] if self.description else []
        parts += [f'docs_url={self.docs_url!r}'] if self.docs_url else []
        parts += [f'docs_description={self.docs_description!r}'] if self.docs_description else []
        return tuple(parts)

    def merge_spec_dict(
            self,
            dependency_to_ref: Mapping['SpecResource', SpecRef],
            present_value: frozendict[str, Jsonable]
    ) -> frozendict[str, Jsonable]:
        return frozendict(present_value | self.get_spec_dict(dependency_to_ref))

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__
